@@ -2,7 +2,7 @@
  * sccsfile.cc: Part of GNU CSSC.
  * 
  * 
- *    Copyright (C) 1997,1998,1999,2001 Free Software Foundation, Inc. 
+ *    Copyright (C) 1997,1998,1999,2001,2003 Free Software Foundation, Inc. 
  * 
  *    This program is free software; you can redistribute it and/or modify
  *    it under the terms of the GNU General Public License as published by
@@ -43,16 +43,56 @@
 #include <unistd.h>		// SEEK_SET on SunOS.
 #endif
 
-#ifdef CONFIG_SCCS_IDS
-static const char rcs_id[] = "CSSC $Id: sccsfile.cc,v 1.48 2002/03/25 00:07:09 james_youngman Exp $";
+#ifdef HAVE_SYS_STAT_H
+#include <sys/stat.h>		/* fstat(), struct stat */
 #endif
 
 
+#ifdef CONFIG_SCCS_IDS
+static const char rcs_id[] = "CSSC $Id: sccsfile.cc,v 1.54 2003/05/29 17:48:25 james_youngman Exp $";
+#endif
+
+#if defined(HAVE_FILENO) && defined(HAVE_FSTAT)
+/* If an SCCS file has a link count greater than one, then the normal 
+ * process of updating the file will break the link.  We try to detect this 
+ * even if the file is being opened to reading only, to give an early
+ * warning (and because SCCS does so).
+ */
+static int just_one_link(FILE *f)
+{
+  int fd = fileno(f);
+  if (fd >= 0)
+    {
+      struct stat st;
+      if (0 != fstat(fd, &st))
+	{
+	  /* We cannot stat the file descriptor.  Perhaps there is a 
+	   * filesystem functionality issue.   If that's the case then we
+	   * will give it the benefit of the doubt on the link coutn front. 
+	   */
+	  return 1;  /* We're happy with the file */
+	}
+      if (st.st_nlink > 1)
+	return 0;		/* We don't like it. */
+    }
+  return 1;			/* OK. */
+}
+#else
+static int just_one_link(FILE *f)
+{
+  /* Without fileno(), we have no way of checking. */
+  return 1;
+}
+#endif
+
 
 /* Static member for opening a SCCS file and then calculating its checksum. */
 #define f f_is_also_a_class_member_variable
 FILE *
-sccs_file::open_sccs_file(const char *name, enum _mode mode, int *sump)
+sccs_file::open_sccs_file(const char *name,
+			  enum _mode mode,
+			  int *sump,
+			  bool* isBKFile)
 {
   FILE *f_local;
 
@@ -64,7 +104,7 @@ sccs_file::open_sccs_file(const char *name, enum _mode mode, int *sump)
   else
     f_local = fopen(name, "r");
 #endif
-  
+
   if (f_local == NULL)
     {
       const char *purpose = (mode == UPDATE) ? "update" : "reading";
@@ -72,8 +112,54 @@ sccs_file::open_sccs_file(const char *name, enum _mode mode, int *sump)
       /*NOTEACHED*/
       return NULL;
     }
-  
-  if (getc(f_local) != '\001' || getc(f_local) != 'h')
+
+  if (!just_one_link(f_local))
+    {
+      // xx: issue error message here
+      errormsg("%s had a hard link count which is greater than one.\n"
+	       "This means that the normal process of updating the file\n"
+	       "would break the hard link.  This error is therefore fatal,\n"
+	       "please fix the problem.\n",
+	       name);
+      (void)fclose(f_local);
+      return NULL;
+    }
+
+  bool badMagic = false;
+  if (getc(f_local) != '\001')
+    {
+      badMagic = true;
+    }
+  else
+    {
+      char magicMarker = getc(f_local);
+      if (magicMarker == 'H')
+	{
+	  if (READ == mode)
+	    {
+	      /* We support read-only access to BK files. */
+	      warning("%s is a BitKeeper file.", name);
+	    }
+	  else
+	    {
+	      errormsg("%s: This is a BitKeeper file, and only read-only "
+		       "access to BitKeeper files is supported at the moment, "
+		       "sorry.\n",
+		       name);
+	      (void)fclose(f_local);
+	      return NULL;
+	    }
+	  // Inform the caller that this is a BK file.
+	  // NB: this is the parameter, not member variable	  
+	  *isBKFile = true;	
+	}
+      else if (magicMarker != 'h')
+	{
+	  badMagic = true;
+	}
+    }
+
+  if (badMagic)
     {
       (void)fclose(f_local);
       s_corrupt_quit("%s: No SCCS-file magic number.  "
@@ -173,9 +259,25 @@ sccs_file::read_line() {
 /* Quits with a message saying that SCCS file is corrupt. */
 
 NORETURN
-sccs_file::corrupt(const char *why) const {
-	s_corrupt_quit("%s: line %d: Corrupted SCCS file. (%s)",
-		       name.c_str(), lineno, why);
+sccs_file::corrupt(const char *fmt, ...) const {
+  char buf[80];
+  const char *p;
+  
+  va_list ap;
+  va_start(ap, fmt);
+  if (-1 == vsnprintf(buf, sizeof(buf), fmt, ap))
+    {
+      warning("%s: error message too long for buffer, so the "
+	      "next message will lack some relevant detail", 
+	      name.c_str());
+      p = fmt;			// best effort
+    }
+  else
+    {
+      p = buf;
+    }
+  s_corrupt_quit("%s: line %d: Corrupted SCCS file. (%s)",
+		 name.c_str(), lineno, p);
 }
 
 
@@ -244,8 +346,8 @@ sccs_file::strict_atoul(const char *s) const
     }
   if (n > limit)
     {
-      errormsg("%s: line %d: Warning: number field exceeds %lu.", 
-	       name.c_str(), lineno, limit);
+      warning("%s: line %d: number field exceeds %lu.", 
+	      name.c_str(), lineno, limit);
     }
   
   return n;
@@ -321,11 +423,11 @@ sccs_file::read_delta() {
 			const char *url = 
 			  "http://cssc.sourceforge.net/"
 			  "documentation/incl_excl.shtml";
-			errormsg("Warning: feature not fully tested: "
-				 "excluded delta in SID %s "
-				 "(see %s for more details)",
-				 tmp.id.as_string().c_str(),
-				 url);
+			warning("feature not fully tested: "
+				"excluded delta in SID %s "
+				"(see %s for more details)",
+				tmp.id.as_string().c_str(),
+				url);
 			tmp.have_excludes = true;
 		      }
 		      break;
@@ -410,11 +512,19 @@ sccs_file::read_delta() {
 		 * to support Larry's extensions, we don't call check_arg()
 		 * here.
 		 */
-		/* check_arg(); */
-		if (bufchar(2) != ' ')
+		if (is_bk_file)
 		  {
-		    saw_unknown_feature("Unknown special comment intro %c%c",
-					c, bufchar(2));
+		    check_bk_comment(c, bufchar(2));
+		  }
+		else
+		  {
+		    check_arg();
+		    if (bufchar(2) != ' ')
+		      {
+			saw_unknown_feature("Unknown special comment "
+					    "intro '%c%c'",
+					    c, bufchar(2));
+		      }
 		  }
 		tmp.comments.add(plinebuf->c_str() + 3);
 	      }
@@ -434,6 +544,76 @@ sccs_file::read_delta() {
 }
 
 
+/* Check for BK flags */
+void
+sccs_file::check_bk_flag(char flagchar) const
+{
+  switch (flagchar)
+    {
+    case 'x':
+      return;
+
+    default:
+      corrupt("Unknown flag '%c'.", flagchar);
+    }
+}
+
+
+/* Find out if it is OK to change the file - called by cdc, rmdel, get -e
+ */
+bool
+sccs_file::edit_mode_ok(bool editing) const
+{
+  if (editing && is_bk_file)
+    {
+      errormsg("%s: This is a BitKeeper file.  Checking BitKeeper files out "
+	       "for editing (or otherwise modifying them) is not supported "
+	       "at the moment, sorry.\n",
+	       name.c_str());
+      return false;
+    }
+  else
+    {
+      return true;
+    }
+}
+
+
+/* Check for BK comments */
+void
+sccs_file::check_bk_comment(char ch, char arg) const
+{
+  ASSERT(is_bk_file);
+  
+  switch (arg)
+    {
+    case 'B':
+    case 'C':
+    case 'F':
+    case 'H':
+    case 'K':
+    case 'M':
+    case 'O':
+    case 'P':
+    case 'R':
+    case 'S':
+    case 'T':
+    case 'V':
+    case 'X':
+    case 'Z':
+      return;
+
+    case ' ': // this is the normal SCCS case.
+      return;
+
+    default:
+      saw_unknown_feature("Unknown special comment intro '%c%c' "
+			  "in BitKeeper file\n",
+			  ch, arg);
+    }
+}
+
+
 /* Seeks on the SCCS file to the start of the body.  This function 
    may be rewritten as fseek() doesn't always work too well on
    text files. */
@@ -474,7 +654,7 @@ sccs_file::get_module_name() const
    locked if it isn't only being read.  */
 
 sccs_file::sccs_file(sccs_name &n, enum _mode m)
-  : name(n), mode(m), lineno(0), xfile_created(false)
+  : name(n), mode(m), lineno(0), xfile_created(false), is_bk_file(false), f(0)
 {
   delta_table = new cssc_delta_table;
   plinebuf     = new cssc_linebuf;
@@ -513,6 +693,7 @@ sccs_file::sccs_file(sccs_name &n, enum _mode m)
   
   if (mode == CREATE)
     {
+      /* f is NULL in this case. */
       return;
     }
 
@@ -520,7 +701,17 @@ sccs_file::sccs_file(sccs_name &n, enum _mode m)
   // a new x-file and then renaming it.   This means that we open
   // the s-file read-only.
   signed int sum = 0;
-  f = open_sccs_file(name.c_str(), READ, &sum);
+  f = open_sccs_file(name.c_str(), READ, &sum, &is_bk_file);
+  
+  if (mode != READ)
+    {
+      if (!edit_mode_ok(true))
+	{
+	  ctor_fail(-1,
+		    "%s: Editing is not supported for BitKeeper files.\n",
+		    name.c_str());
+	}
+    }
   
   /* open_sccs_file() returns normally if everything went OK, or if 
    * there was an IO error on an apparently valid file.  If this is 
@@ -532,12 +723,29 @@ sccs_file::sccs_file(sccs_name &n, enum _mode m)
     }
   
   int c = read_line();
-  ASSERT(c == 'h');
+  
+  // open_sccs_file() should have already checked that the first line
+  // is ^Ah or ^Ah, so this assertion is really just checking that
+  // open_sccs_file() did the right thing.
+  if (is_bk_file)
+    {
+      ASSERT(c == 'H');
+    }
+  else
+    {
+      ASSERT(c == 'h');
+    }
 
   /* the checksum is represented in the file as decimal.
    */
   signed int given_sum = 0;
-  if (1 != sscanf(plinebuf->c_str(), "%*ch%d", &given_sum))
+  const char *format;
+  if (is_bk_file)
+    format = "%*cH%d";
+  else
+    format = "%*ch%d";
+  
+  if (1 != sscanf(plinebuf->c_str(), format, &given_sum))
     {
       errormsg("Expected checksum line, found line beginning '%.3s'\n",
 	       plinebuf->c_str());
@@ -557,9 +765,9 @@ sccs_file::sccs_file(sccs_name &n, enum _mode m)
 	    }
 	  else
 	    {
-	      errormsg("%s: Warning: bad checksum "
-		       "(expected=%d, calculated %d).\n",
-		       name.c_str(), given_sum, sum);
+	      warning("%s: bad checksum "
+		      "(expected=%d, calculated %d).\n",
+		      name.c_str(), given_sum, sum);
 	    }
 	}
     }
@@ -704,11 +912,18 @@ sccs_file::sccs_file(sccs_name &n, enum _mode m)
 	else if (got_arg && '0' == *arg)
 	  flags.encoded = 0;
 	else
-	  corrupt("Bad value for 'e' flag.");
+	  corrupt("Bad value '%c' for 'e' flag.", arg[0]);
 	break;
 	
       default:
-	corrupt("Unknown flag.");
+	if (is_bk_file)
+	  {
+	    check_bk_flag(bufchar(3));
+	  }
+	else
+	  {
+	    corrupt("Unknown flag '%c'.", bufchar(3));
+	  }
       }
       
       c = read_line();
@@ -953,7 +1168,7 @@ bool sccs_file::branches_allowed() const
  * checks what mode we're using the SCCS file in, and reacts
  * accordingly.
  */
-void sccs_file::saw_unknown_feature(const char *fmt, ...)
+void sccs_file::saw_unknown_feature(const char *fmt, ...) const
 {
   va_list ap;
 
